import copy
import math
import sys
import numpy as np
from contextlib import closing
from io import StringIO
from gym import utils
from gym.envs.toy_text import discrete
from src.rlpe.Environments.ApplePicking.apple_picking_constants import *

# MAP = [
#     "+---------+",
#     "|A: |z| :A|",
#     "| : : : : |",
#     "| : : : : |",
#     "| : : : : |",
#     "|S|z|A|z|A|",
#     "+---------+",
# ]

MAP = [
    "+---------+",
    "|A: : : :A|",
    "| : : : : |",
    "| : : : : |",
    "| : : : : |",
    "|S: :A: :A|",
    "+---------+",
]


# MAP = [
#     "+-----+",
#     "|A: :A|",
#     "| : :A|",
#     "|S: :A|",
#     "+-----+",
# ]


def try_step_south_or_east(place, max_place):
    new_place = min(place + 1, max_place)
    return new_place


def try_step_west_or_north(place, max_place):
    new_place = max(place - 1, max_place)
    return new_place


# def apples_valid(a1, a2, a3, a4):
#     result = not (
#             a1 != 0 and a1 == a2 and a1 == a3 and a1 == a4 and a2 != 0 and a2 == a3 and a2 == a4 and a3 != 0 and a3 == a4)
#     tamp_apple_arr = [a1, a2, a3, a4]
#     max_apple = max([a1, a2, a3, a4])
#     for i in range(1, max_apple):
#         if i not in tamp_apple_arr:
#             result = False
#     return result


class ApplePickingEnv(discrete.DiscreteEnv):
    """
    Actions:
    There are 6 discrete deterministic actions:
    - 0: move south
    - 1: move north
    - 2: move east
    - 3: move west
    - 4: pickup an apple
    Rewards:
    - (-1): step
    - (-3): entering thorny area
    - (-10): excecute "pick
    Rendering:
    - Purple A: not collected apple
    - White A: collected apple
    - Green z: thorn
    - Blue: the collector
    state space is represented by:
        (collector_row, collector_col, apple1_location, apple2_location, apple3_location, apple4_location)
    """
    metadata = {'render.modes': ['human', 'ansi']}

    def __init__(self):
        self.desc = np.asarray(MAP, dtype='c')
        w, h = self.desc.shape
        self.last_action = None
        self.apple_locations, self.thorny_locations, self.start_position = self.get_info_from_map()
        self.original_apple_locations = copy.deepcopy(self.apple_locations)
        self.num_of_picked_apples = 0
        self.num_rows = int(w - 2)
        self.num_columns = int((h - 1) / 2)
        self.num_states = (self.num_rows * self.num_columns) * int(math.pow(2, len(self.apple_locations)))
        self.max_row = self.num_rows - 1
        self.max_col = self.num_columns - 1
        self.initial_state_distribution = np.zeros(self.num_states)
        self.num_actions = len(ACTIONS)
        self.P = self.build_transition_matrix()
        # self.initial_state_distribution /= self.initial_state_distribution.sum()
        discrete.DiscreteEnv.__init__(self, self.num_states, self.num_actions, self.P, self.initial_state_distribution)

    def build_transition_matrix(self):
        """
        Build the transition matrix.
        You can work with deterministic environment or stochastic one by changing the flag in the arguments.
        return: dictionary with the transition matrix
        """
        P = {state: {action: [] for action in range(self.num_actions)} for state in range(self.num_states)}
        for row in range(self.num_rows):
            for col in range(self.num_columns):
                for apple1 in range(2):
                    for apple2 in range(2):
                        for apple3 in range(2):
                            for apple4 in range(2):
                                state = self.encode(row, col, apple1, apple2, apple3, apple4)
                                apple_arr = [apple1, apple2, apple3, apple4]
                                num_of_picked_apples = sum(apple_arr)
                                if row == self.start_position[0] and col == self.start_position[
                                    1] and num_of_picked_apples == 0:
                                    self.initial_state_distribution[state] += 1.0
                                for action in range(self.num_actions):
                                    new_row, new_col = row, col
                                    new_apple1, new_apple2, new_apple3, new_apple4 = apple1, apple2, apple3, apple4
                                    new_apple_arr = [new_apple1, new_apple2, new_apple3, new_apple4]
                                    reward = REWARD_DICT[STEP_REWARD]  # default reward when there is no pickup
                                    done = True if num_of_picked_apples == 4 else False
                                    collector_loc = (row, col)

                                    if action in [SOUTH, NORTH, WEST, EAST]:
                                        new_row, new_col, reward = self.try_to_move(action, row, col)

                                    elif action == PICKUP:
                                        new_apple_arr, reward, done = self.try_picking_up(collector_loc, done,
                                                                                          new_apple_arr, reward)
                                    self.num_of_picked_apples = num_of_picked_apples
                                    new_state = self.encode(new_row, new_col, *new_apple_arr)
                                    if self.near_thorny_wall(row, col, ABOVE):
                                        P[state][action] = self.get_stochastic_probs(action, row, col, apple_arr,
                                                                                     new_state, reward, done, ABOVE)
                                    elif self.near_thorny_wall(row, col, UNDER):
                                        P[state][action] = self.get_stochastic_probs(action, row, col, apple_arr,
                                                                                     new_state, reward, done, UNDER)
                                    else:
                                        P[state][action] = self.get_stochastic_probs(action, row, col, apple_arr,
                                                                                     new_state, reward, done)
        return P

    def encode(self, collector_row, collector_col, apple1, apple2, apple3, apple4):
        # (5), 5, 5, 5, 5, 5
        # (num_rows), num_columns, 2 X 4
        i = collector_row
        i *= self.num_columns
        i += collector_col
        i *= 2
        i += apple1
        i *= 2
        i += apple2
        i *= 2
        i += apple3
        i *= 2
        i += apple4
        return i

    def decode(self, i):
        # 2 X 4, 5, (5)
        # 2 X 4, num_columns, (num_rows)
        out = []
        out.append(i % 2)
        i = i // 2
        out.append(i % 2)
        i = i // 2
        out.append(i % 2)
        i = i // 2
        out.append(i % 2)
        i = i // 2
        out.append(i % self.num_columns)
        i = i // self.num_columns
        out.append(i)
        assert 0 <= i < self.num_rows
        return list(reversed(out))

    def near_thorny_wall(self, new_row, new_col, value):
        for thorny_wall in self.thorny_locations:
            if new_row == thorny_wall[0] + value and new_col == thorny_wall[1]:
                return True
        return False

    def check_if_state_is_legal(self, state, return_idxes=False):
        if isinstance(state, int):
            collector_row, collector_col, apple1, apple2, apple3, apple4 = self.decode(state)
        else:
            collector_row, collector_col, apple1, apple2, apple3, apple4 = state
        not_valid_idx = []
        if collector_row < 0 or collector_row >= self.num_rows:
            not_valid_idx.append(0)
        if collector_col < 0 or collector_col >= self.num_columns:
            not_valid_idx.append(1)
        if apple_locations < 0 or apple_locations > len(self.apple_locations):  # TODO - bug!!!
            not_valid_idx.append(2)
        state_is_legal = (len(not_valid_idx) == 0)
        if return_idxes:
            return state_is_legal, not_valid_idx
        return state_is_legal

    def render(self, mode='human'):
        outfile = StringIO() if mode == 'ansi' else sys.stdout

        out = self.desc.copy().tolist()
        out = [[c.decode('utf-8') for c in line] for line in out]
        collector_row, collector_col, apple1, apple2, apple3, apple4 = self.decode(self.s)
        apple_arr = apple1, apple2, apple3, apple4

        def ul(x):
            return "_" if x == " " else x

        def colorize_loc(j, cur_color, apple_arr):
            if apple_arr[j] == NOT_COLLECTED:
                xc, yc = self.apple_locations[j]
                out[1 + xc][2 * yc + 1] = utils.colorize(out[1 + xc][2 * yc + 1], cur_color)

        out[1 + collector_row][2 * collector_col + 1] = utils.colorize(out[1 + collector_row][2 * collector_col + 1],
                                                                       'blue', highlight=True)

        for (di, dj) in self.thorny_locations:
            out[1 + di][2 * dj + 1] = utils.colorize(ul(out[1 + di][2 * dj + 1]), 'green', bold=True)

        # colors = ['magenta', 'cyan', 'crimson', 'yellow', 'red', 'white']
        for j in range(len(self.apple_locations)):
            colorize_loc(j, 'magenta', apple_arr)

        outfile.write("\n".join(["".join(row) for row in out]) + "\n")
        if self.lastaction is not None:
            outfile.write("  ({})\n".format(["South", "North", "East", "West", "Pickup"][self.lastaction]))
        else:
            outfile.write("\n")
        print("current state: ", self.decode(self.s), ", last action: ", self.last_action)
        # No need to return anything for human
        if mode != 'human':
            with closing(outfile):
                return outfile.getvalue()

    def reset(self):
        self.s = discrete.categorical_sample(self.isd, self.np_random)
        self.last_action = None
        return int(self.s)

    def step(self, a):
        transitions = self.P[self.s][a]
        i = discrete.categorical_sample([t[0] if len(t) > 0 else 0 for t in transitions], self.np_random)
        p, s, r, d = transitions[i]
        self.s = s
        self.last_action = a
        return int(s), r, d, {"prob": p}

    def try_to_move(self, action, row, col):
        new_row, new_col = row, col
        if action == SOUTH:
            new_row = try_step_south_or_east(row, self.max_row)
        elif action == NORTH:
            new_row = try_step_west_or_north(row, 0)
        elif action == EAST and self.no_wall_to_the_right(row, col):
            new_col = try_step_south_or_east(col, self.max_col)
        elif action == WEST and self.no_wall_to_the_left(row, col):
            new_col = try_step_west_or_north(col, 0)
        reward = REWARD_DICT[STEP_INTO_THORN_REWARD] if (new_row, new_col) in self.thorny_locations else REWARD_DICT[
            STEP_REWARD]
        return new_row, new_col, reward

    def get_stochastic_probs(self, action, row, col, apple_arr, new_state, reward, done, near_thorny=0):
        # return [(DETERMINISTIC_PROB, new_state, reward, done)]
        if near_thorny == 0:
            return self.prob_list_for_no_move_action(action, new_state, reward, done)
        risky_action = None
        prob_list = [tuple() for _ in range(self.num_actions)]
        if near_thorny == ABOVE:
            risky_action = SOUTH
        elif near_thorny == UNDER:
            risky_action = NORTH
        action_prob = (STOCHASTIC_PROB, new_state, reward, done)
        prob_list[action] = action_prob
        for prob_act in range(len(prob_list)):
            if prob_act != action:
                if prob_act == risky_action:
                    new_row, new_col, reward = self.try_to_move(prob_act, row, col)
                    done = False
                    new_state = self.encode(new_row, new_col, *apple_arr)
                    prob_list[prob_act] = (STOCHASTIC_PROB_THORNY, new_state, reward, done)
                else:
                    prob_list[prob_act] = (0.0, new_state, reward, done)
            elif action == risky_action:
                prob_list[prob_act] = (1.0, new_state, reward, done)
        return prob_list

    def prob_list_for_no_move_action(self, action, new_state, reward, done):
        prob_list = [tuple() for _ in range(self.num_actions)]
        prob_list[action] = (DETERMINISTIC_PROB, new_state, reward, done)
        for a in range(len(prob_list)):
            if a != action:
                prob_list[a] = (0.0, new_state, reward, done)
        return prob_list

    def no_wall_to_the_right(self, row, col):
        return self.desc[1 + row, 2 * col + 2] == b":"

    def no_wall_to_the_left(self, row, col):
        return self.desc[1 + row, 2 * col] == b":"

    def try_picking_up(self, collector_loc, done, new_apple_arr, reward):
        for i, apple_loc in enumerate(self.apple_locations):
            if collector_loc == apple_loc and new_apple_arr[i] == NOT_COLLECTED:
                old_apple_arr = copy.deepcopy(new_apple_arr)
                new_apple_arr[i] = COLLECTED
                if sum(new_apple_arr) == 4 and sum(old_apple_arr) != 4:
                    done = True
                reward = REWARD_DICT[APPLE_PICKUP_REWARD]
                break
        if reward != REWARD_DICT[APPLE_PICKUP_REWARD]:
            reward = REWARD_DICT[BAD_APPLE_PICKUP_REWARD]
        return new_apple_arr, reward, done

    def get_info_from_map(self):
        apple_locations, thorny_wall_locations, start_position = [], [], None
        h, w = self.desc.shape
        h, w = (h - 2), (w - 2)
        for x in range(1, h + 1):
            for y in range(1, w + 1):
                c = self.desc[x][y]
                if c == b'A':
                    apple_locations.append((x - 1, int(y / 2)))
                elif c == b'z':
                    thorny_wall_locations.append((x - 1, int(y / 2)))
                elif c == b'S':
                    start_position = (x - 1, int(y / 2))

        return apple_locations, thorny_wall_locations, start_position


# if __name__ == '__main__':
#     new_env = ApplePickingEnv()
#     new_env.reset()
#     actions = [1, 1, 1, 1, 4, 2, 0, 0, 2, 2, 1, 1, 2, 4, 0, 0, 0, 0, 4, 1, 1, 3, 3, 0, 0, 4, 1]
#     actions = [1, 1, 1, 1, 4, 2, 0, 2, 2, 1, 1, 2, 4, 0, 0, 0, 0, 4, 1, 1, 3, 3, 0, 0, 4, 1]
#     all_reward = 0
#     for act in actions:
#         new_env.render()
#         next_s, r, d, prob = new_env.step(act)
#         all_reward += r
#         print("state:", new_env.decode(next_s))
#         print(f"{next_s}, {r}, {d}, {prob}")
#         print("all_reward:", all_reward)
#         if d:
#             break



import sys
import numpy as np
from contextlib import closing
from io import StringIO
from gym import utils
from gym.envs.toy_text import discrete
from src.rlpe.Environments.SingleTaxiEnv.single_taxi_constants import *

MAP = [
    "+---------+",
    "|R: | : :G|",
    "| : | : : |",
    "| : : : : |",
    "| : : | : |",
    "|Y: :F| :B|",
    "+---------+",
]


def try_step_south_or_east(fuel, place, max_place):
    new_place = min(place + 1, max_place)
    if new_place == place + 1:
        fuel -= 1
    return fuel, new_place


def try_step_west_or_north(fuel, place, max_place):
    new_place = max(place - 1, max_place)
    if new_place == place - 1:
        fuel -= 1
    return fuel, new_place


class SingleTaxiEnv(discrete.DiscreteEnv):
    """
    The Taxi Problem
    from "Hierarchical Reinforcement Learning with the MAXQ Value Function Decomposition"
    by Tom Dietterich
    Description:
    There are four designated locations in the grid world indicated by R(ed), G(reen), Y(ellow), and B(lue). When the episode starts, the taxi starts off at a random square and the passenger is at a random location. The taxi drives to the passenger's location, picks up the passenger, drives to the passenger's destination (another one of the four specified locations), and then drops off the passenger. Once the passenger is dropped off, the episode ends.
    Observations:
    There are 500 discrete states since there are 25 taxi positions, 5 possible locations of the passenger (including the case when the passenger is in the taxi), and 4 destination locations.
    Passenger locations:
    - 0: R(ed)
    - 1: G(reen)
    - 2: Y(ellow)
    - 3: B(lue)
    - 4: in taxi
    Destinations:
    - 0: R(ed)
    - 1: G(reen)
    - 2: Y(ellow)
    - 3: B(lue)
    Actions:
    There are 6 discrete deterministic actions:
    - 0: move south
    - 1: move north
    - 2: move east
    - 3: move west
    - 4: pickup passenger
    - 5: drop off passenger
    - 6: refuel the taxi
    Rewards:
    There is a default per-step reward of -1,
    except for delivering the passenger, which is +20,
    or executing "pickup" and "drop-off" actions illegally, which is -10.
    Rendering:
    - blue: passenger
    - magenta: destination
    - yellow: empty taxi
    - green: full taxi
    - other letters (R, G, Y and B): locations for passengers and destinations
    state space is represented by:
        (taxi_row, taxi_col, passenger_location, destination)
    """
    metadata = {'render.modes': ['human', 'ansi']}

    def __init__(self, deterministic=True):
        self.init_state, self.init_row, self.init_col = 20154, 4, 0
        self.deterministic = deterministic
        self.desc = np.asarray(MAP, dtype='c')
        w, h = self.desc.shape
        self.last_action = None
        self.passengers_locations, self.fuel_station = self.get_info_from_map()
        self.taxi_fuel = MAX_FUEL
        self.num_rows = int(w - 2)
        self.num_columns = int((h - 1) / 2)
        self.num_states = (self.num_rows * self.num_columns) * (len(self.passengers_locations) + 1) * (
            len(self.passengers_locations)) * MAX_FUEL
        self.max_row = self.num_rows - 1
        self.max_col = self.num_columns - 1
        self.initial_state_distribution = np.zeros(self.num_states)
        self.num_actions = len(ACTIONS)
        self.passenger_in_taxi = len(self.passengers_locations)
        self.P = self.build_transition_matrix(deterministic=deterministic)
        # self.initial_state_distribution /= self.initial_state_distribution.sum()
        discrete.DiscreteEnv.__init__(self, self.num_states, self.num_actions, self.P, self.initial_state_distribution)

    def build_transition_matrix(self, deterministic=True):
        """
        Build the transition matrix.
        You can work with deterministic environment or stochastic one by changing the flag in the arguments.
        return: dictionary with the transition matrix
        """
        P = {state: {action: [] for action in range(self.num_actions)} for state in range(self.num_states)}
        for row in range(self.num_rows):
            for col in range(self.num_columns):
                for pass_idx in range(len(self.passengers_locations) + 1):  # +1 for being inside taxi
                    for dest_idx in range(len(self.passengers_locations)):
                        for fuel in range(self.taxi_fuel):
                            init_fuel = fuel
                            state = self.encode(row, col, pass_idx, dest_idx, fuel)
                            if self.is_possible_initial_state(pass_idx, dest_idx, row,
                                                              col) and state == self.init_state:
                                self.initial_state_distribution[state] += 1.0
                            for action in range(self.num_actions):
                                new_row, new_col, new_pass_idx = row, col, pass_idx
                                reward = REWARD_DICT[STEP_REWARD]  # default reward when there is no pickup/dropoff
                                done = False
                                taxi_loc = (row, col)

                                if fuel == 0:
                                    done = True
                                    reward = REWARD_DICT[NO_FUEL_REWARD]
                                else:
                                    if action in [SOUTH, NORTH, WEST, EAST]:
                                        new_row, new_col, fuel = self.try_to_move(action, fuel, row, col)

                                    elif action == PICKUP:
                                        new_pass_idx, reward = self.try_picking_up(pass_idx, taxi_loc, reward,
                                                                                   new_pass_idx)
                                    elif action == DROPOFF:
                                        new_pass_idx, reward, done = self.try_dropping_off(taxi_loc, dest_idx, pass_idx,
                                                                                           new_pass_idx, reward, done)
                                    elif action == REFUEL:
                                        fuel, reward = self.try_to_refuel(taxi_loc, fuel)
                                new_state = self.encode(new_row, new_col, new_pass_idx, dest_idx, fuel)
                                if deterministic:
                                    P[state][action].append((DETERMINISTIC_PROB, new_state, reward, done))
                                else:
                                    probs = self.get_stochastic_probs(action, row, col, pass_idx, dest_idx, init_fuel,
                                                                      new_state, reward, done)
                                    P[state][action] = probs
                                fuel = init_fuel
        return P

    def encode(self, taxi_row, taxi_col, pass_loc, dest_idx, fuel):
        # (5), 5, 5, 4, 50
        # (num_rows), num_columns, len(passengers_locations) + 1, len(passengers_locations), MAX_FUEL
        i = taxi_row
        i *= self.num_columns
        i += taxi_col
        i *= (len(self.passengers_locations) + 1)  # +1 for in taxi location
        i += pass_loc
        i *= len(self.passengers_locations)
        i += dest_idx
        i *= MAX_FUEL
        i += fuel
        return i

    def decode(self, i):
        # 50, 4, 5, 5, (5)
        # MAX_FUEL, len(passengers_locations), len(passengers_locations) + 1, num_columns, (num_rows)
        out = []
        out.append(i % MAX_FUEL)
        i = i // MAX_FUEL
        out.append(i % len(self.passengers_locations))
        i = i // len(self.passengers_locations)
        out.append(i % (len(self.passengers_locations) + 1))
        i = i // (len(self.passengers_locations) + 1)  # +1 for in taxi location
        out.append(i % self.num_columns)
        i = i // self.num_columns
        out.append(i)
        assert 0 <= i < self.num_rows
        return list(reversed(out))

    def check_if_state_is_legal(self, state, return_idxes=False):
        taxi_row, taxi_col, pass_loc, dest_idx, fuel = state
        not_valid_idx = []
        if taxi_row < 0 or taxi_row >= self.num_rows:
            not_valid_idx.append(0)
        if taxi_col < 0 or taxi_col >= self.num_columns:
            not_valid_idx.append(1)
        if pass_loc < 0 or pass_loc > len(self.passengers_locations):
            not_valid_idx.append(2)
        if dest_idx < 0 or dest_idx >= len(self.passengers_locations):
            not_valid_idx.append(3)
        if fuel < 0 or fuel > MAX_FUEL:
            not_valid_idx.append(4)
        state_is_legal = (len(not_valid_idx) == 0)
        if return_idxes:
            return state_is_legal, not_valid_idx
        return state_is_legal

    def render(self, mode='human'):
        outfile = StringIO() if mode == 'ansi' else sys.stdout

        out = self.desc.copy().tolist()
        out = [[c.decode('utf-8') for c in line] for line in out]
        taxi_row, taxi_col, pass_idx, dest_idx, fuel = self.decode(self.s)

        def ul(x):
            return "_" if x == " " else x

        if pass_idx < 4:
            out[1 + taxi_row][2 * taxi_col + 1] = utils.colorize(out[1 + taxi_row][2 * taxi_col + 1], 'yellow',
                                                                 highlight=True)
            pi, pj = self.passengers_locations[pass_idx]
            out[1 + pi][2 * pj + 1] = utils.colorize(out[1 + pi][2 * pj + 1], 'blue', bold=True)
        else:  # passenger in taxi
            out[1 + taxi_row][2 * taxi_col + 1] = utils.colorize(ul(out[1 + taxi_row][2 * taxi_col + 1]), 'green',
                                                                 highlight=True)

        di, dj = self.passengers_locations[dest_idx]
        out[1 + di][2 * dj + 1] = utils.colorize(out[1 + di][2 * dj + 1], 'magenta')
        outfile.write("\n".join(["".join(row) for row in out]) + "\n")
        if self.lastaction is not None:
            outfile.write("  ({})\n".format(["South", "North", "East", "West", "Pickup", "Dropoff"][self.lastaction]))
        else:
            outfile.write("\n")
        print("current state: ", self.decode(self.s), ", last action: ", self.last_action)
        # No need to return anything for human
        if mode != 'human':
            with closing(outfile):
                return outfile.getvalue()

    def reset(self):
        self.s = discrete.categorical_sample(self.isd, self.np_random)
        self.last_action = None
        return int(self.s)

    def step(self, a):
        transitions = self.P[self.s][a]
        i = discrete.categorical_sample([t[0] if len(t) > 0 else 0 for t in transitions], self.np_random)
        p, s, r, d = transitions[i]
        self.s = s
        self.last_action = a
        return int(s), r, d, {"prob": p}

    def try_to_move(self, action, fuel, row, col):
        new_row, new_col = row, col
        if action == SOUTH and fuel != 0:
            fuel, new_row = try_step_south_or_east(fuel, row, self.max_row)
        elif action == NORTH and fuel != 0:
            fuel, new_row = try_step_west_or_north(fuel, row, 0)
        elif action == EAST and self.no_wall_to_the_right(row, col) and fuel != 0:
            fuel, new_col = try_step_south_or_east(fuel, col, self.max_col)
        elif action == WEST and self.no_wall_to_the_left(row, col) and fuel != 0:
            fuel, new_col = try_step_west_or_north(fuel, col, 0)
        return new_row, new_col, fuel

    def get_stochastic_probs(self, action, row, col, pass_idx, dest_idx, fuel, new_state, reward, done):
        if action in [REFUEL, PICKUP, DROPOFF]:
            return self.prob_list_for_no_move_action(action, new_state, reward, done)
        prob_list = [tuple() for _ in range(self.num_actions)]
        action_prob = (STOCHASTIC_PROB, new_state, reward, done)
        prob_list[action] = action_prob
        for prob_act in range(len(prob_list)):
            init_fuel = fuel
            if prob_act != action and prob_act < 4:
                if fuel == 0:
                    new_row, new_col, reward, done = row, col, REWARD_DICT[NO_FUEL_REWARD], True
                else:
                    new_row, new_col, fuel = self.try_to_move(prob_act, fuel, row, col)
                    reward, done = REWARD_DICT[STEP_REWARD], False
                new_state = self.encode(new_row, new_col, pass_idx, dest_idx, fuel)
                prob_list[prob_act] = (STOCHASTIC_PROB_OTHER_ACTIONS, new_state, reward, done)
            fuel = init_fuel
        return prob_list

    def prob_list_for_no_move_action(self, action, new_state, reward, done):
        prob_list = [tuple() for _ in range(self.num_actions)]
        prob_list[action] = (DETERMINISTIC_PROB, new_state, reward, done)
        return prob_list

    def no_wall_to_the_right(self, row, col):
        return self.desc[1 + row, 2 * col + 2] == b":"

    def no_wall_to_the_left(self, row, col):
        return self.desc[1 + row, 2 * col] == b":"

    def try_picking_up(self, pass_idx, taxi_loc, reward, new_pass_idx):
        if pass_idx < self.passenger_in_taxi and taxi_loc == self.passengers_locations[pass_idx]:
            new_pass_idx = self.passenger_in_taxi
        else:  # passenger not at location
            reward = REWARD_DICT[BAD_PICKUP_REWARD]
        return new_pass_idx, reward

    def try_dropping_off(self, taxi_loc, dest_idx, pass_idx, new_pass_idx, reward, done):
        if (taxi_loc == self.passengers_locations[dest_idx]) and pass_idx == self.passenger_in_taxi:
            new_pass_idx = dest_idx
            done = True
            reward = REWARD_DICT[DROPOFF_REWARD]
        elif (taxi_loc in self.passengers_locations) and pass_idx == self.passenger_in_taxi:
            new_pass_idx = self.passengers_locations.index(taxi_loc)
        else:  # dropoff at wrong location
            reward = REWARD_DICT[BAD_DROPOFF_REWARD]
        return new_pass_idx, reward, done

    def try_to_refuel(self, taxi_loc, fuel):
        if taxi_loc == self.fuel_station and fuel < (MAX_FUEL - 10):
            reward = REWARD_DICT[REFUEL_REWARD]
            fuel = MAX_FUEL - 1
        else:
            reward = REWARD_DICT[BAD_REFUEL_REWARD]
        return fuel, reward

    def is_possible_initial_state(self, pass_idx, dest_idx, row, col):
        return pass_idx < 4 and pass_idx != dest_idx

    def get_info_from_map(self):
        fuel_station = None
        passenger_locations = []
        h, w = self.desc.shape
        h, w = (h - 2), (w - 2)
        for x in range(1, h + 1):
            for y in range(1, w + 1):
                c = self.desc[x][y]
                if c == b'R' or c == b'G' or c == b'B' or c == b'Y':
                    passenger_locations.append((x - 1, int(y / 2)))
                elif c == b'F':
                    fuel_station = (x - 1, int(y / 2))
        return passenger_locations, fuel_station

#
# if __name__ == '__main__':
#     new_env = SingleTaxiEnv(deterministic=True)
#     new_env.reset()
#     actions = [2, 6, 6, 6, 6, 6, 3, 1, 1, 1, 4, 0, 2, 2, 0, 0, 5]
#   actions = [2, 2, 6, 1, 1, 3, 3, 1, 1, 4, 0, 0, 2, 2, 2, 2, 0, 0, 5]
#     all_reward = 0
#     for act in actions:
#         new_env.render()
#         next_s, r, d, prob = new_env.step(act)
#         all_reward += r
#         print("state:", new_env.decode(next_s))
#         print("reward:", r, "done:", d, "prob:", prob)
#         print("all_reward:", all_reward)
#
#     passenger_locations, fuel_station = new_env.get_info_from_map()
# new_env.reset()
# actions = [2, 2, 6, 1, 1, 3, 3, 1, 1, 4, 0, 0, 2, 2, 2, 2, 0, 0, 5]
# all_reward = 0
# for act in actions:
#     new_env.render()
#     next_s, r, d, prob = new_env.step(act)
#     all_reward += r
#     print("state:", new_env.decode(next_s))
#     print("reward:", r, "done:", d, "prob:", prob)
#     print("all_reward:", all_reward)



import copy
import itertools
import os
import pickle
import re
from src.rlpe.Agents.RL_agents import rl_agent
from src.rlpe.Transforms.transform_constants import SINGLE_TAXI_EXAMPLE
from src.rlpe.constants import *
import warnings


def make_dir(dir_name):
    try:
        os.mkdir(dir_name)
    except OSError:
        print("Directory %s exists already" % dir_name)


def save_pkl_file(file_name, data):
    if file_name[-4:] != ".pkl":
        file_name = file_name + ".pkl"
    file = open(file_name, 'wb')
    pickle.dump(data, file)
    file.close()


def load_pkl_file(file_path):
    if file_path[-4:] != ".pkl":
        file_path = file_path + ".pkl"
    try:
        a_file = open(file_path, "rb")
        obj = pickle.load(a_file)
    except:
        warnings.warn(f"ERROR: cannot load object from file path: {file_path}")
        obj = None
    return obj


def save_trained_model(agent, agent_name, transform_name):
    model_name = agent_name + '_' + transform_name
    dir_path = TRAINED_AGENT_SAVE_PATH
    make_dir(dir_path + model_name)
    dir_path = dir_path + model_name + '/'
    agent.model.save_weights(dir_path + model_name)


def make_or_restore_model(env, agent_name, transform_name):
    # Prepare a directory to store all the checkpoints.
    checkpoint_dir_path = TRAINED_AGENT_SAVE_PATH + f"{agent_name}_{transform_name}/"
    if not os.path.exists(checkpoint_dir_path):
        os.makedirs(checkpoint_dir_path)
    # Either restore the latest model, or create a fresh one if there is no checkpoint available.
    checkpoints = [checkpoint_dir_path + name for name in os.listdir(checkpoint_dir_path)]
    restored = True if checkpoints else False
    if checkpoints:
        latest_checkpoint = max(checkpoints, key=os.path.getctime)
        print("Restoring from", latest_checkpoint)

        return load_existing_agent(env, agent_name, transform_name), restored
    print("Creating a new model")
    return rl_agent.create_agent(env, agent_name), restored


def load_existing_results(agent_name, env_name, num_episodes=ITER_NUM, dir_path=TRAINED_AGENT_RESULTS_PATH):
    info_name = f"{env_name}_{agent_name}"
    file_path = dir_path + info_name + f"_{num_episodes}/"
    try:
        explanation = load_pkl_file(file_path + info_name + "_explanation.pkl")
        result = load_pkl_file(file_path + info_name + "_result.pkl")
    except:
        warnings.warn(f"ERROR: Not available results for env_name: {env_name}, agent_name: {agent_name}")
        result, explanation = [], []

    return result, explanation


def load_env_preconditions(env_name):
    if env_name == SINGLE_TAXI_EXAMPLE:
        a_file = open(PRECONDITIONS_PATH, "rb")
        preconditions = pickle.load(a_file)
        return preconditions
    else:
        raise Exception("not valid env_name")


def load_existing_transforms_from_dir(working_dir=TRANSFORMS_PATH):
    possible_env_files = os.listdir(working_dir)
    cur_transforms = dict()
    dict_idx = 0
    for i, file_name in enumerate(possible_env_files):
        transform_name, new_env = load_transform_by_name(file_name, dir_name=working_dir)
        cur_transforms[dict_idx] = transform_name, new_env
        dict_idx += 1
    global transforms
    transforms = cur_transforms
    return cur_transforms


def get_precondition_actions_from_string(pre_string):
    string_for_extracting_actions = copy.deepcopy(pre_string)
    string_for_extracting_actions = re.sub("([\(\[]).*?([\)\]])", "\g<1>\g<2>", string_for_extracting_actions)
    precondition_actions = [int(s) for s in string_for_extracting_actions.split('_') if s.isdigit()]
    return precondition_actions


def get_precondition_idx_from_string(single_pre_string):
    pre_idx = single_pre_string[single_pre_string.find("(") + 1:single_pre_string.find(")")]
    pre_idx = tuple(int(pre) for pre in pre_idx if pre != ' ' and pre != ',')
    return pre_idx


def get_precondition_val_from_string(single_pre_string):
    pre_val = single_pre_string[single_pre_string.find("[") + 1:single_pre_string.find("]")]
    pre_val = [int(pre) for pre in pre_val if pre != ' ']
    return pre_val


def load_existing_transforms_by_anticipated_policy(env_name, anticipated_policy, working_dir=TRANSFORMS_PATH):
    possible_env_files = os.listdir(working_dir)
    cur_transforms = dict()
    dict_idx = -1
    anticipate_policy_actions = anticipated_policy.values()
    all_possible_anticipated_policies = list(
        itertools.combinations(anticipate_policy_actions, len(anticipate_policy_actions)))
    for i, file_name in enumerate(possible_env_files):
        precondition_actions = get_precondition_actions_from_string(file_name)
        for policy_action_list in all_possible_anticipated_policies:
            policy_action_list = [p for tmp in policy_action_list for p in tmp]
            is_precondition_actions_relevant = any(
                pre_action in policy_action_list for pre_action in precondition_actions)
            if is_precondition_actions_relevant:
                dict_idx += 1
                # precondition_idx = file_name[file_name.find("(") + 1:file_name.find(")")]
                # precondition_val = file_name[file_name.find("[") + 1:file_name.find("]")]
                transform_name, new_env = load_transform_by_name(file_name)
                cur_transforms[dict_idx] = transform_name, new_env
    global transforms
    transforms = cur_transforms
    return cur_transforms


def load_transform_by_name(file_name, dir_name=TRANSFORMS_PATH):
    if not file_name[:-4] == ".pkl":
        file_name += ".pkl"
    transform_name = os.path.basename(file_name)[:-4]
    try:
        file = open(dir_name + file_name, "rb")
        new_env = pickle.load(file)
    except:
        warnings.warn(f"[save_load_utils][load_transform_by_name]: could not load existing transform")
        new_env = None
    return transform_name, new_env


def load_existing_agent(env, agent_name, env_name, trained_agents_path=TRAINED_AGENT_SAVE_PATH):
    model_name = agent_name + '_' + env_name
    dir_path = trained_agents_path + model_name + '/'
    new_agent = rl_agent.create_agent(env, agent_name)
    try:
        new_agent = new_agent.load_existing_agent(dir_path + model_name)
    except:
        warnings.warn(f"ERROR: bad agent: {dir_path + model_name}")
        new_agent = None
    return new_agent



from src.rlpe.Observer.lunar_lander_expert import LunarLanderExpert
from src.rlpe.Observer.single_taxi_expert import SingleTaxiExpert
from src.rlpe.Observer.taxi_expert import Taxi_Expert
from src.rlpe.Transforms.transform_constants import *
from src.rlpe.save_load_utils import *

transforms = []
transform_index = -1


def get_env(env_name, number_of_agents=1):
    """
    :param env_name:
    :param number_of_agents:
    :return:
    """
    if env_name == TAXI:
        from src.rlpe.Environments.taxi_environment_wrapper import TaxiSimpleEnv
        return TaxiSimpleEnv()
    elif env_name == TAXI_EXAMPLE:
        from src.rlpe.Environments.taxi_environment_wrapper import TaxiSimpleExampleEnv
        return TaxiSimpleExampleEnv()
    elif env_name == SINGLE_TAXI_EXAMPLE:
        from src.rlpe.Environments.SingleTaxiEnv.single_taxi_wrapper import SingleTaxiSimpleEnv
        return SingleTaxiSimpleEnv()
    elif env_name == SINGLE_FROZEN_EXAMPLE:
        from src.rlpe.Environments.frozenlake_environment import FrozenLakeEnv
        return FrozenLakeEnv()
    elif env_name == LUNAR_LANDER:
        from src.rlpe.Environments.lunar_lander_wrapper import LunarLenderWrapper
        return LunarLenderWrapper()
    elif env_name == SEARCH_TRANSFORM_TAXI_ENV:
        from src.rlpe.Environments.SingleTaxiEnv.single_taxi_wrapper import SingleTaxiSimpleEnv
        new_env = load_pkl_file(TRANSFORM_SEARCH_TAXI_ENV_PATH)
        return new_env
    elif env_name == APPLE_PICKING:
        from src.rlpe.Environments.ApplePicking.apple_picking_env import ApplePickingEnv
        return ApplePickingEnv()
    elif env_name == SPEAKER_LISTENER:
        from supersuit import pad_observations_v0, pad_action_space_v0
        from pettingzoo.mpe import simple_speaker_listener_v3
        from ray.tune.registry import register_env
        from ray.rllib.env import PettingZooEnv

        def create_env(args):
            env = simple_speaker_listener_v3.env()
            env = pad_action_space_v0(env)
            env = pad_observations_v0(env)
            return env

        get_env_lambda = lambda config: PettingZooEnv(create_env(config))
        register_env(SPEAKER_LISTENER, lambda config: get_env_lambda(config))
        return get_env_lambda({}), SPEAKER_LISTENER


def get_expert(env_name, env):
    if env_name == TAXI_EXAMPLE:
        return Taxi_Expert(env)
    elif env_name == SINGLE_TAXI_EXAMPLE:
        return SingleTaxiExpert(env)
    elif env_name == LUNAR_LANDER:
        return LunarLanderExpert(env)


def is_partial_obs_equal_to_state(partial_obs, state):
    if len(partial_obs) != len(state):
        raise Exception("The length of the partial observation differs from the state")
    for i in range(len(partial_obs)):
        if partial_obs[i] is None:
            continue
        if partial_obs[i] != state[i]:
            return False
    return True


def is_actions_align(action, target_policy_actions):
    target_policy_actions = [target_policy_actions] if isinstance(target_policy_actions, int) else target_policy_actions
    for act in target_policy_actions:
        if act == action:
            return True
    return False


def print_num_of_success_failed_policies(num_of_success_policies, num_of_failed_policies):
    print("num_of_success_policies: ", num_of_success_policies)
    print("num_of_failed_policies: ", num_of_failed_policies)


# def is_anticipated_policy_achieved_old(env, agent, anticipated_policy):
#     agent.evaluating = True
#     num_of_success_policies, num_of_failed_policies = 0, 0
#     for partial_obs in anticipated_policy.keys():
#         original_partial_obs = partial_obs
#         partial_obs = list(partial_obs)
#         states_from_partial_obs = env.get_states_from_partial_obs(partial_obs)
#         for i, state in enumerate(states_from_partial_obs):
#             action = agent.compute_action(state)
#             if is_actions_align(action, anticipated_policy[original_partial_obs]):
#                 num_of_success_policies += 1
#             else:
#                 num_of_failed_policies += 1
#
#     all_policies = num_of_success_policies + num_of_failed_policies
#     success_rate = num_of_success_policies / (all_policies if all_policies != 0 else 1)
#     print("\nSuccess rate:", success_rate)
#     agent.evaluating = False
#     return success_rate > 0.8, success_rate

def add_one_if_in_dict(given_dict, key):
    given_dict[key] = given_dict[key] + 1 if key in given_dict.keys() else 1
    return given_dict


def map_action(transformed_env, agent_action):
    real_action = -1
    transform_name = ""
    for transform_name, abstract_action in transformed_env.translation_action_to_precondition_dict.items():
        if agent_action == abstract_action:
            real_action = int(transform_name[0])
            break
    return real_action, transform_name


def map_actions_to_explanation(original_env, agent, search_taxi_env, anticipated_policy):
    explanation = dict()
    agent.evaluating = True
    cur_state = search_taxi_env.reset()
    done, steps_num = False, 0
    while not done and steps_num < 100:
        agent_action = agent.compute_action(cur_state)
        if agent_action not in [_ for _ in range(original_env.num_actions)]:
            agent_real_action, transform_name = map_action(search_taxi_env, agent_action)
            explanation[transform_name] = (agent_real_action, agent_action)
        cur_state, reward, done, prob = search_taxi_env.step(agent_action)
        steps_num += 1
    return explanation


    # for partial_obs in anticipated_policy.keys():
    #     # original_partial_obs = partial_obs
    #     partial_obs = list(partial_obs)
    #     states_from_partial_obs = original_env.get_states_from_partial_obs(partial_obs)
    #     for i, state in enumerate(states_from_partial_obs):
    #         agent_action = agent.compute_action(state)
    #         if agent_action not in [_ for _ in range(original_env.num_actions)]:
    #             agent_real_action, transform_name = map_action(search_taxi_env, agent_action)
    #             explanation[transform_name] = (agent_real_action, agent_action)
    #             agent_action = agent_real_action
    # agent.evaluating = False
    # return explanation


def is_anticipated_policy_achieved(original_env, agent, anticipated_policy, transformed_env=None):
    agent.evaluating = True
    success_policies_set, failed_policies_set, not_reached_policy_set = set(), set(), set()
    cur_state = original_env.reset()
    done, steps_num = False, 0
    while not done and steps_num < 100:
        agent_action = agent.compute_action(cur_state)
        if transformed_env:
            if agent_action not in [_ for _ in range(original_env.num_actions)]:
                agent_action, _ = map_action(transformed_env, agent_action)
        is_align, anticipated_action, anticipated_state = is_state_align_with_anticipated_policy(original_env,
                                                                                                 cur_state,
                                                                                                 anticipated_policy)
        if is_align:
            if is_actions_align(agent_action, anticipated_action):
                success_policies_set.add(anticipated_state)
            else:
                failed_policies_set.add(anticipated_state)
        cur_state, reward, done, prob = original_env.step(agent_action)
        steps_num += 1
    for anticipated_state in anticipated_policy.keys():
        if anticipated_state not in success_policies_set and anticipated_state not in failed_policies_set:
            not_reached_policy_set.add(anticipated_state)
    num_of_success_policies, num_of_failed_policies = len(success_policies_set), len(failed_policies_set)
    all_policies = num_of_success_policies + num_of_failed_policies
    success_rate = num_of_success_policies / (all_policies if all_policies != 0 else 1)
    print("\nSuccess rate:", success_rate)
    if len(not_reached_policy_set) != 0:
        print(f"There are some states that the agent can't reach: {not_reached_policy_set}")
    agent.evaluating = False
    return success_rate == 1.0, success_rate


def is_state_align_with_anticipated_state(env, state, anticipated_state):
    anticipated_state = list(anticipated_state)
    decoded_state = env.decode(state)
    for (anticipated_feature, feature) in zip(anticipated_state, decoded_state):
        if anticipated_feature is not None and feature != anticipated_feature:
            return False
    return True


def is_state_align_with_anticipated_policy(env, state, anticipated_policy):
    for anticipated_state, anticipated_action in anticipated_policy.items():
        if is_state_align_with_anticipated_state(env, state, anticipated_state):
            return True, anticipated_action[0], anticipated_state
    return False, None, None


def get_transformed_env(env_name):
    if env_name == TAXI_EXAMPLE:
        from src.rlpe.Transforms.taxi_transforms import TaxiTransformedEnv
        return TaxiTransformedEnv
    elif env_name == SINGLE_TAXI_EXAMPLE:
        from src.rlpe.Transforms.single_taxi_transforms import SingleTaxiTransformedEnv
        return SingleTaxiTransformedEnv
    elif env_name == LUNAR_LANDER:
        from src.rlpe.Transforms.lunar_lander_transforms import LunarLanderTransformedEnv
        return LunarLanderTransformedEnv


# def set_all_possible_transforms_old(original_env, env_name):
#     binary_permutations = ["".join(seq) for seq in product("01", repeat=original_env.transform_num)]
#     transforms = {}
#     for per in binary_permutations:
#         bool_params = tuple(True if int(dig) == 1 else False for dig in per)
#         if any(bool_params):
#             transformed_env = get_transformed_env(env_name)
#             transform_name = get_transform_name(env_name, bool_params)
#             transforms[bool_params] = (transform_name, transformed_env)
#     return transforms

def set_all_possible_transforms(original_env, env_name, anticipated_policy):
    env_preconditions = load_env_preconditions(env_name)
    basic_relevant_transforms = dict()
    for state, actions in anticipated_policy.items():
        for action in actions:
            if action not in basic_relevant_transforms:
                basic_relevant_transforms[action] = env_preconditions.not_allowed_features[action]
    preconditions_num = 0
    for action, precondition in basic_relevant_transforms.items():
        preconditions_num += sum([len(precondition[idx]) for idx in precondition.keys()])
    global transforms
    transforms = basic_relevant_transforms
    return basic_relevant_transforms


def get_next_transformed_env():
    global transforms, transform_index
    transform_index += 1
    yield transforms[transform_index]


def get_transform_name(env_name, bool_params):
    if env_name == TAXI_EXAMPLE:
        from src.rlpe.Transforms.taxi_transforms import get_taxi_transform_name
        return get_taxi_transform_name(bool_params)
    elif env_name == SINGLE_TAXI_EXAMPLE:
        from src.rlpe.Transforms.single_taxi_transforms import get_single_taxi_transform_name
        return get_single_taxi_transform_name(bool_params)
    elif env_name == LUNAR_LANDER:
        from src.rlpe.Transforms.lunar_lander_transforms import get_lunar_lander_transform_name
        return get_lunar_lander_transform_name(bool_params)


def is_transform_actions_influence_anticipated_policy(anticipated_actions, transform_actions):
    influence = True
    for act in transform_actions:
        if not influence:
            break
        influence = False
        for anti_act in anticipated_actions:
            if act in anti_act:
                influence = True
    return influence



from src.rlpe.Agents.RL_agents.rl_agent import *
import src.rlpe.Agents.RL_agents.rl_agent as rl_agent
from src.rlpe.Agents.abstract_agent import AbstractAgent
import numpy as np
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Embedding, Reshape
from tensorflow.keras.optimizers import Adam
from rl.agents.dqn import DQNAgent
from rl.policy import EpsGreedyQPolicy
from rl.memory import SequentialMemory
from src.rlpe.constants import *
# import tensorflow as tf

HANDS_ON_DQN = "hands_on_dqn"
KERAS_DQN = "keras_dqn"
Q_LEARNING = "q_learning"
KERAS_SARSA = "keras_sarsa"
KERAS_CEM = "cem_sarsa"
VALUE_ITERATION = "value_iteration"

MAX_EXPLORATION_RATE = 1
MIN_EXPLORATION_RATE = 0.01
EXPLORATION_DECAY_RATE = 0.001

TERMINAL_STATE = 'TERMINAL_STATE'


def flip_coin(p):
    r = random.random()
    return r < p


class QLearningAgent(AbstractAgent):
    def __init__(self, env, epsilon=1, alpha=0.7, gamma=0.6, timesteps_per_episode=60):
        """
        alpha    - learning rate
        epsilon  - exploration rate
        gamma    - discount factor
        num_actions - number of actions in the current environment
        """
        super().__init__(env, timesteps_per_episode)
        self.episodesSoFar = 0
        self.accumTrainRewards = 0.0
        self.accumTestRewards = 0.0
        self.num_training = 500
        self.evaluating = False
        """ Parameters """
        self.alpha = alpha
        self.gamma = gamma

        self.num_actions = env.action_space.n
        self.epsilon = epsilon
        self.q_values = {}
        self.terminal_states = None
        self.episodeRewards = 0
        self.policy_dict = {}

    def get_q_value(self, state, action):
        """
          Returns Q(state,action) or 0.0 if we never seen a state or (state,action) tuple
        """
        if (state, action) in self.q_values:
            return self.q_values[(state, action)]
        return 0.0

    def get_policy(self, state):
        """
          Computes the best action to take in a state.
        """
        actions = [i for i in range(self.env.action_space.n)]
        q_value_dict = {action: self.get_q_value(state, action) for action in actions}
        max_action = max(q_value_dict, key=q_value_dict.get)
        return max_action

    def compute_action(self, state):
        """
          Computes the action to take in the current state.  With
          probability self.epsilon, we should take a random action and
          take the best policy action otherwise.
          Should use transform_fn if it exist.
        """
        actions = [i for i in range(self.env.action_space.n)]
        if flip_coin(self.epsilon) and not self.evaluating:
            return random.choice(actions)
        max_action = self.get_policy(state)
        return max_action

    def episode_callback(self, state, action, reward, next_state, terminated):
        self.update_alpha()
        self.update_q_values(state, action, reward, next_state)
        return next_state

    def update_q_values(self, state, action, reward, next_state):
        old_value = self.get_q_value(state, action)
        next_max = np.max(self.get_policy(next_state))

        new_value = (1 - self.alpha) * old_value + self.alpha * (reward + self.gamma * next_max)
        self.q_values[(state, action)] = new_value

    def update_alpha(self):
        """
        Updates the exploration rate in the end of each episode.
        """
        self.epsilon = MIN_EXPLORATION_RATE + (MAX_EXPLORATION_RATE - MIN_EXPLORATION_RATE) * np.exp(
            -EXPLORATION_DECAY_RATE * self.episodesSoFar)  # Exploration rate decay

    def start_episode(self):
        self.episodeRewards = 0.0

    def stop_episode(self):
        """
          Called by environment when episode is done
        """
        if not self.evaluating:
            self.accumTrainRewards += self.episodeRewards
        else:
            self.accumTestRewards += self.episodeRewards
        self.episodesSoFar += 1
        if self.evaluating:
            # Take off the training wheels
            self.epsilon = 0.0  # no exploration
            self.alpha = 0.0  # no learning
        self.start_episode()

    def get_legal_actions(self, state):
        if self.num_actions == 6 or self.num_actions == 4:
            return [i for i in range(self.num_actions)]
        if state == TERMINAL_STATE:
            return ()
        elif state in self.terminal_states:
            return ('exit',)
        return 'up', 'left', 'down', 'right'

    def set_terminal_states(self, terminal_states):
        self.terminal_states = terminal_states

    def run(self):
        result = rl_agent.run_episode(self.env, self)
        self.episodeRewards = result[TOTAL_EPISODE_REWARD]
        return result

    def evaluate(self, visualize=False):
        # print("================ DISPLAY ====================")
        self.evaluating = True
        result = rl_agent.run_episode(self.env, self, method=EVALUATE)
        self.evaluating = False
        return result


class HandsOnDQNAgent(AbstractAgent):
    def __init__(self, env, timesteps_per_episode=200, batch_size=32):
        super().__init__(env, timesteps_per_episode)
        self.evaluating = False
        self.batch_size = batch_size
        # Initialize attributes
        self._state_size = env.num_states
        self._action_size = env.action_space.n
        self._optimizer = Adam(learning_rate=0.01)

        self.experience_replay = deque(maxlen=2000)

        # Initialize discount and exploration rate
        self.gamma = 0.6
        self.epsilon = 0.1

        # Build networks
        self.q_network = self._build_compile_model()
        self.target_network = self._build_compile_model()
        self.align_target_model()

    def store(self, state, action, reward, next_state, terminated):
        self.experience_replay.append((state, action, reward, next_state, terminated))

    def _build_compile_model(self):
        model = Sequential()
        model.add(Embedding(self._state_size, 10, input_length=1))
        model.add(Reshape((10,)))
        model.add(Dense(50, activation='relu', input_shape=(1,)))
        model.add(Dense(50, activation='relu'))
        model.add(Dense(self._action_size, activation='linear'))

        model.compile(loss='mse', optimizer=self._optimizer)
        return model

    def align_target_model(self):
        self.target_network.set_weights(self.q_network.get_weights())

    def compute_action(self, state):
        # state = state[0] if len(state) == 1 else state
        if np.random.rand() <= self.epsilon:
            return self.env.action_space.sample()

        q_values = self.q_network.predict(state)
        return np.argmax(q_values[0])

    def episode_callback(self, state, action, reward, next_state, terminated):
        # next_state = np.reshape(next_state, [1, 1])
        self.store(state, action, reward, next_state, terminated)

        if len(self.experience_replay) > self.batch_size:
            self.retrain(self.batch_size)

        return next_state

    def stop_episode(self):
        if self.evaluating:
            # Take off the training wheels
            self.epsilon = 0.0  # no exploration
        self.align_target_model()

    def run(self):
        result = rl_agent.run_episode(self.env, self)
        return result

    def retrain(self, batch_size):
        minibatch = random.sample(self.experience_replay, batch_size)

        for state, action, reward, next_state, terminated in minibatch:

            target = self.q_network.predict(state)

            if terminated:
                target[0][action] = reward
            else:
                t = self.target_network.predict(next_state)
                target[0][action] = reward + self.gamma * np.amax(t)

            self.q_network.fit(state, target, epochs=1, verbose=0)

    def evaluate(self):
        # print("================ DISPLAY ====================")
        self.evaluating = True
        result = rl_agent.run_episode(self.env, self, method=EVALUATE)
        return result


class DQNKeras(AbstractAgent):
    def __init__(self, env, callbacks=None, timesteps_per_episode=60, batch_size=32):
        super().__init__(env, timesteps_per_episode)
        self.action_size = env.action_space.n
        self.state_size = env.num_states
        self.callbacks = callbacks
        np.random.seed(123)
        if hasattr(env, '_seed'):
            env._seed(123)
        # Build networks
        self.model = self._build_compile_model()
        memory = SequentialMemory(limit=50000, window_length=1)
        policy = EpsGreedyQPolicy()
        self.dqn_only_embedding = DQNAgent(model=self.model, nb_actions=self.action_size, memory=memory,
                                           nb_steps_warmup=500,
                                           target_model_update=1e-2, policy=policy)

    def _build_compile_model(self):
        model = Sequential()
        model.add(Embedding(self.state_size, 10, input_length=1))  # 600000
        model.add(Reshape((10,)))
        # model.add(Flatten())
        model.add(Dense(50, activation='relu'))
        model.add(Dense(50, activation='relu'))
        model.add(Dense(50, activation='relu'))
        model.add(Dense(self.action_size, activation='linear'))
        # print(model.summary())
        return model

    def run(self) -> {str: float}:
        """
        The agent's training method.
        Returns: a dictionary - {"episode_reward_mean": __, "episode_reward_min": __, "episode_reward_max": __,
        "episode_len_mean": __}
        """
        self.dqn_only_embedding.compile(Adam(lr=1e-3), metrics=['mae'])
        history = self.dqn_only_embedding.fit(self.env, nb_steps=ITER_NUM, visualize=False, verbose=1,
                                              nb_max_episode_steps=ITER_NUM, log_interval=10000)

        # history = self.dqn_only_embedding.fit(self.env, nb_steps=2000, visualize=False, verbose=1,
        #                                       nb_max_episode_steps=60, log_interval=2000)
        result = {EPISODE_REWARD_MEAN: np.array(history.history["episode_reward"]),
                  EPISODE_STEP_NUM_MEAN: np.array(history.history["nb_episode_steps"]),
                  EPISODE_REWARD_MIN: np.empty([]),
                  EPISODE_REWARD_MAX: np.empty([]), EPISODE_VARIANCE: np.empty([])}
        return result

    def compute_action(self, state) -> int:
        """
        Computes the best action from a given state.
        Returns: a int that represents the best action.
        """
        # self.epsilon *= self.epsilon_decay
        # self.epsilon = max(self.epsilon_min, self.epsilon)
        # if np.random.random() < self.epsilon:
        #     return self.env.action_space.sample()
        state = np.array([[state]])
        return int(np.argmax(self.model.predict(state)))

    def stop_episode(self):
        pass

    def episode_callback(self, state, action, reward, next_state, terminated):
        pass

    def evaluate(self, visualize=True):
        self.dqn_only_embedding.test(self.env, nb_episodes=1, visualize=visualize, nb_max_episode_steps=70)

    def load_existing_agent(self, dir_path):
        self.model.load_weights(dir_path)
        self.dqn_only_embedding.compile(Adam(lr=1e-3), metrics=['mae'])
        return self



TAXIS_LOC_IDX = 0
FUELS_IDX = 1
PASS_START_LOC_IDX = 2
PASS_DEST_IDX = 3
PASS_STATUS_IDX = 4

TAXI = "taxi"
TAXI_EXAMPLE = "taxi_example"
SINGLE_TAXI_EXAMPLE = "single_taxi_example"
SINGLE_FROZEN_EXAMPLE = "single_frozens_example"

LUNAR_LANDER = "lunar_lander"
DIM_REDUCTION_IDX = PASS_STATUS_IDX
ENV_NAME = "taxi"

TAXI_LOC_X = "TX_"
TAXI_LOC_Y = "TY_"
FUEL = "F_"
PASS_LOC = "PL_"
PASS_DEST = "PD_"
PASS_LOC_X = "PX_"
PASS_LOC_Y = "PY_"
PASS_DEST_X = "DX_"
PASS_DEST_Y = "DY_"
PASS_STATUS = "S_"
WALLS = "W_"
REWARD = "R_"
STATE_VISIBILITY_TRANSFORM = "SVT"
ALL_OUTCOME_DETERMINIZATION = "AOD_"
MOST_LIKELY_OUTCOME = "MLO_"
PRECONDITION_RELAXATION = "PR_"

X_COORDINATE = "XC_"
Y_COORDINATE = "YC_"
X_VELOCITY = "XV_"
Y_VELOCITY = "YV_"
ORIENTATION = "OR_"
ANGULAR_VELOCITY = "AV_"






from src.rlpe.train_agent import *
from src.rlpe.utils import is_transform_actions_influence_anticipated_policy


class TransformNode:
    def __init__(self, actions, idxes, values, val, depth, parents, preconditions_graph):
        self.val = val
        self.depth = depth
        self.actions = actions
        self.idxes = idxes
        self.values = values
        self.parents = parents
        self.adjacent_nodes = []
        self.transform_name = ""
        if not preconditions_graph:
            self.transform_name = ORIGINAL_ENV
        else:
            for a, i, v in zip(self.actions, self.idxes, self.values):
                self.transform_name += f"{a}_{i}_{v}_"
            self.transform_name = self.transform_name[:-1]
        self.visited = False
        if preconditions_graph:
            preconditions_graph.graph[self.transform_name] = self


# This class represents a directed graph using adjacency list representation
class PreconditionsGraph:
    def __init__(self, env_name, preconditions, anticipated_policy):
        self.root = TransformNode(actions=[], idxes=[], values=[], val=0, depth=0, parents=[],
                                  preconditions_graph=None)
        self.anticipated_policy = anticipated_policy
        self.queue = [self.root]
        self.root.visited = True
        self.graph = defaultdict(TransformNode)
        self.graph[ORIGINAL_ENV] = self.root
        self.max_depth = len(anticipated_policy) + 1
        self.build_preconditions_graph(preconditions, anticipated_policy)
        self.original_env = get_env(env_name)

    def add_edge(self, u, v):
        self.graph[u].adjacent_nodes.append(v)

    def bfs(self):
        self.original_env = get_env(SINGLE_TAXI_EXAMPLE)  # TODO - Temporary!! to delete!!
        satisfaction = False
        explanation_env = None
        while self.queue and not satisfaction:
            transform_node = self.queue.pop(0)
            print(transform_node.transform_name, end=" ")
            for neighbour_mdp in transform_node.adjacent_nodes:
                if not neighbour_mdp.visited:
                    self.queue.append(neighbour_mdp)
                    neighbour_mdp.visited = True
                    satisfaction = train_agent_and_save_results(self.original_env, neighbour_mdp)
                    if satisfaction:
                        explanation_env = neighbour_mdp
                        break
        if satisfaction:
            print(f"Congratulations!! We have an explanation env for you! Its name: {explanation_env.transform_name}")
        else:
            print("Sorry, we don't know why the agent doing weird stuff :-(")

    def get_graph_node_by_name(self, node_name):
        return self.graph[node_name]

    def update_graph(self, all_preconditions_list, depth, precondition_node):
        possible_parents = all_preconditions_list[depth - 1]
        pre_actions_possible_parents = [get_precondition_actions_from_string(pre_string[0][0])[0] for pre_string in
                                        possible_parents]
        for i, action in enumerate(pre_actions_possible_parents):
            if action in precondition_node.actions:
                parent_node = self.get_graph_node_by_name(possible_parents[i][0])
                precondition_node.parents.append(parent_node)
                parent_node.adjacent_nodes.append(precondition_node)

    def build_preconditions_graph(self, preconditions, anticipated_policy):
        basic_preconditions_list = []
        depth = 1
        for action, action_info in preconditions.items():
            for idx, values in action_info.items():
                for val in values:
                    pre_string = f"{action}_{idx}_{val}"
                    precondition_actions = get_precondition_actions_from_string(pre_string)
                    influence = is_transform_actions_influence_anticipated_policy(list(anticipated_policy.values()),
                                                                                  precondition_actions)
                    if influence:
                        basic_preconditions_list.append(f"{action}_{idx}_{val}")
                        node = TransformNode(actions=[action], idxes=[idx], values=[val], val=0, depth=depth,
                                             parents=[self.root], preconditions_graph=self)
                        self.root.adjacent_nodes.append(node)

        all_preconditions_list = {1: tuple(tuple([pre]) for pre in basic_preconditions_list)}
        for i in range(2, self.max_depth):
            depth += 1
            print(f"depth: {depth}")
            all_preconditions_list[i] = tuple(itertools.combinations(basic_preconditions_list, i))
            for pre in all_preconditions_list[i]:
                actions, idxes, vals = list(), tuple(), list()
                idxes, vals = list(), list()
                for p in pre:
                    action = get_precondition_actions_from_string(p)
                    idx = get_precondition_idx_from_string(p)
                    val = get_precondition_val_from_string(p)
                    actions += action
                    idxes.append(idx)
                    vals.append(val)
                node = TransformNode(actions=actions, idxes=idxes, values=vals, val=0, depth=depth, parents=[],
                                     preconditions_graph=self)
                self.update_graph(all_preconditions_list, depth, node)


def get_precondition_from_transform_node(mdp_node):
    precondition = {}
    for act, idx, val in zip(mdp_node.actions, mdp_node.idxes, mdp_node.values):
        if act in precondition.keys():
            if idx in precondition[act].keys():
                precondition[act][idx].append(val)
            else:
                precondition[act][idx] = [val]
        else:
            precondition[act] = {idx: val}
    return precondition


def train_agent_and_save_results(original_env, mdp_node):
    # transform_name, new_env = load_transform_by_name(neighbour_mdp.transform_name + ".pkl")
    precondition = get_precondition_from_transform_node(mdp_node)
    new_env = generate_transformed_env(precondition, env_file_name=TRANSFORMS_PATH + mdp_node.transform_name, save=True,
                                       try_load_existing=True)
    satisfaction = generate_agent(original_env, KERAS_DQN, ITER_NUM, new_env, mdp_node.transform_name)
    return satisfaction

# if __name__ == '__main__':
#     from Transforms import env_precinditions
#
#     cur_env_preconditions = load_pkl_file("../" + TAXI_TRANSFORM_DATA_PATH + "small_taxi_env_preconditions.pkl")
#     # precondition_graph = PreconditionsGraph(cur_env_preconditions.not_allowed_features, ANTICIPATED_POLICY)
#     precondition_graph = load_pkl_file("precondition_graph_small_taxi_env.pkl")
#     save_pkl_file("precondition_graph_small_taxi_env.pkl", precondition_graph)
#     precondition_graph.bfs()



import gym
from src.rlpe.Environments.MultiTaxiEnv.multitaxienv.taxi_environment import TaxiEnv
from itertools import product
import numpy as np
from src.rlpe.Environments.abstract_wrapper_env import AbstractWrapperEnv

NEW_MAP = [
    "+---------+",
    "|X: | : :X|",
    "| : | : : |",
    "| : : : : |",
    "| : : | : |",
    "|X: :F| :X|",
    "+---------+",
]
TAXI_NAME = "taxi_1"
TAXI_INIT_POS = [4, 0]
PASS_INIT_POS = [0, 0]
PASS_DEST = [4, 4]

# ========================================== DO NOT DELETE ========================================== #
# This part is for running changing environments:
# all_possible_envs = list(product([0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 4], [0, 4], [0, 4], [0, 4]))
# all_possible_envs = [list(s) for s in all_possible_envs if (s[2] != s[4] or (s[3] != s[5]))]
# global env_idx
# env_idx = -1
# in the reset function:
# self.taxis_fixed_locations = [[all_possible_envs[env_idx][0], all_possible_envs[env_idx][1]]]
# self.passengers_start_fixed_locations = [[all_possible_envs[env_idx][2], all_possible_envs[env_idx][3]]]
# self.passengers_fixed_destinations = [[all_possible_envs[env_idx][4], all_possible_envs[env_idx][5]]]
# self.passengers_locations = [self.taxis_fixed_locations[0], self.passengers_fixed_destinations[0]]
# ========================================== DO NOT DELETE ========================================== #

def set_up_env_idx():
    global env_idx
    env_idx += 1


class TaxiSimpleEnv(TaxiEnv, AbstractWrapperEnv):
    def __init__(self, _=0, num_taxis: int = 1, num_passengers: int = 1, max_fuel: list = None,
                 domain_map: list = None, taxis_capacity: list = None, collision_sensitive_domain: bool = True,
                 fuel_type_list: list = None, option_to_stand_by: bool = False):
        # Initializing default values
        self.transform_num = 10
        self.num_taxis = num_taxis
        if max_fuel is None:
            self.max_fuel = [100] * num_taxis
        else:
            self.max_fuel = max_fuel

        if domain_map is None:
            self.desc = np.asarray(NEW_MAP, dtype='c')
        else:
            self.desc = np.asarray(domain_map, dtype='c')

        if taxis_capacity is None:
            self.taxis_capacity = [1] * num_passengers
        else:
            self.taxis_capacity = taxis_capacity

        if fuel_type_list is None:
            self.fuel_type_list = ['F'] * num_passengers
        else:
            self.fuel_type_list = fuel_type_list

        # Relevant features for map orientation, notice that we can only drive between the columns (':')
        self.num_rows = num_rows = len(self.desc) - 2
        self.num_columns = num_columns = len(self.desc[0][1:-1:2])

        # Set locations of passengers and fuel stations according to the map.
        self.passengers_locations = []
        self.fuel_station1 = None
        self.fuel_station2 = None
        self.fuel_stations = []

        # initializing map with passengers and fuel stations
        for i, row in enumerate(self.desc[1:-1]):
            for j, char in enumerate(row[1:-1:2]):
                loc = [i, j]
                if char == b'X':
                    self.passengers_locations.append(loc)
                elif char == b'F':
                    self.fuel_station1 = loc
                    self.fuel_stations.append(loc)
                elif char == b'G':
                    self.fuel_station2 = loc
                    self.fuel_stations.append(loc)

        self.coordinates = [[i, j] for i in range(num_rows) for j in range(num_columns)]

        # self.num_taxis = num_taxis
        self.taxis_names = ["taxi_" + str(index + 1) for index in range(self.num_taxis)]

        self.collision_sensitive_domain = collision_sensitive_domain

        # Indicator list of 1's (collided) and 0's (not-collided) of all taxis
        self.collided = np.zeros(self.num_taxis)

        self.option_to_standby = option_to_stand_by

        # A list to indicate whether the engine of taxi i is on (1) or off (0), all taxis start as on.
        self.engine_status_list = list(np.ones(self.num_taxis).astype(bool))

        self.num_passengers = num_passengers

        # Available actions in relation to all actions based on environment parameters.
        self.available_actions_indexes, self.index_action_dictionary, self.action_index_dictionary \
            = self._set_available_actions_dictionary()
        self.num_actions = len(self.available_actions_indexes)
        self.action_space = gym.spaces.Discrete(self.num_actions)
        self.observation_space = gym.spaces.MultiDiscrete(self._get_observation_space_list())
        self.bounded = False

        self.last_action = None
        self.num_states = self._get_num_states()

        self._seed()
        self.state = None
        self.dones = {taxi_name: False for taxi_name in self.taxis_names}
        self.dones['__all__'] = False

        self.np_random = None

        self.taxis_fixed_locations = [TAXI_INIT_POS]
        self.passengers_start_fixed_locations = [PASS_INIT_POS]
        self.passengers_fixed_destinations = [PASS_DEST]
        self.passengers_locations = [PASS_INIT_POS, PASS_DEST]

        self.reset()

    def reset(self):
        obs = super(TaxiSimpleEnv, self).reset()[TAXI_NAME]
        encoded_state = self.encode(obs)
        return encoded_state

    def step(self, action):
        action_dict = {TAXI_NAME: action}
        next_state, reward, dones, info = super(TaxiSimpleEnv, self).step(action_dict)
        next_state, reward, dones = next_state[TAXI_NAME][0], reward[TAXI_NAME], dones[TAXI_NAME]
        encoded_next_state = self.encode(next_state)
        return encoded_next_state, reward, dones, info

    def encode(self, state):
        # (self.num_rows), self.num_columns, max_fuel[0] + 1, self.num_rows, self.num_columns, self.passengers_locations, 4
        taxi_row, taxi_col, fuel, pass_loc_x, pass_loc_y, dest_idx_x, dest_idx_y, pass_status = state[0], state[1], \
                                                                                                state[2], state[3], \
                                                                                                state[4], state[5], \
                                                                                                state[6], state[7]
        # dest_idx = self._get_pass_dest_idx(dest_idx_x, dest_idx_y)

        i = taxi_row

        i *= self.num_columns
        i += taxi_col

        i *= (self.max_fuel[0] + 1)
        i += fuel

        i *= self.num_rows
        i += pass_loc_x

        i *= self.num_columns
        i += pass_loc_y

        # i *= len(self.passengers_fixed_destinations)
        # i += dest_idx

        i *= 3
        i += pass_status
        return i

    def decode(self, i):
        # 4, self.passengers_locations, self.num_columns, self.num_rows, max_fuel[0] + 1, self.num_columns, self.num_rows
        j = i
        out = []

        passenger_status = [(i % 3)]
        out.append(passenger_status)
        i = i // 3

        # passenger_dest_idx = [self.passengers_fixed_destinations[i % len(self.passengers_fixed_destinations)]]
        out.append([PASS_DEST])
        # i = i // len(self.passengers_fixed_destinations)

        passenger_loc_y = i % self.num_columns
        i = i // self.num_columns
        passenger_loc_x = i % self.num_rows
        i = i // self.num_rows
        passenger_location = [[passenger_loc_x, passenger_loc_y]]
        out.append(passenger_location)

        fuel = [i % (self.max_fuel[0] + 1)]
        out.append(fuel)
        i = i // (self.max_fuel[0] + 1)

        taxi_y = i % self.num_columns
        i = i // self.num_columns
        taxi_x = i
        taxi_loc = [[taxi_x, taxi_y]]
        out.append(taxi_loc)

        assert 0 <= i < self.num_rows

        return list(reversed(out))

    def _get_pass_dest_idx(self, dest_idx_x, dest_idx_y):
        dest_idx = -1
        for i, loc in enumerate(self.passengers_locations):
            if dest_idx_x == loc[0] and dest_idx_y == loc[1]:
                dest_idx = i
                break
        if dest_idx == -1:
            raise Exception("no such destination!")
        return dest_idx

    def flatten_state(self, state):
        taxi_loc, fuel, pas_loc, pas_des, status = state[0][0], state[1], state[2][0], state[3][0], state[4]
        return taxi_loc + fuel + pas_loc + pas_des + status

    def get_states_from_partial_obs(self, partial_obs):
        partial_obs_aligned_with_env = False
        iter_num = 200
        while not partial_obs_aligned_with_env and iter_num != 0:
            obs = self.reset()
            obs = self.flatten_state(self.decode(obs))
            if self._is_aligned(obs, partial_obs):
                partial_obs_aligned_with_env = True
            iter_num -= 1

        if partial_obs_aligned_with_env:
            taxi_x = [partial_obs[0]] if (partial_obs[0] is not None) else list(range(self.num_columns))
            taxi_y = [partial_obs[1]] if (partial_obs[1] is not None) else list(range(self.num_rows))
            fuel = [partial_obs[2]] if partial_obs[2] else list(range(1, self.max_fuel[0]))
            passenger_start_x, passenger_start_y = [obs[3]], [obs[4]]
            passenger_dest_x, passenger_dest_y = [obs[5]], [obs[6]]
            passenger_status = [partial_obs[7]] if partial_obs[7] else list(range(1, 4))
            states = list(
                product(taxi_x, taxi_y, fuel, passenger_start_x, passenger_start_y, passenger_dest_x, passenger_dest_y,
                        passenger_status, repeat=1))
            states = [self.encode(state) for state in states]
        else:
            states = []
        return states

    def _is_aligned(self, obs, partial_obs):
        passenger_start_x, passenger_start_y, passenger_dest_x, passenger_dest_y = self._get_passenger_info(partial_obs)
        return (passenger_start_x is None or passenger_start_x == obs[3]) and (
                passenger_start_y is None or passenger_start_y == obs[4]) and (
                       passenger_dest_x is None or passenger_dest_x == obs[5]) and (
                       passenger_dest_y is None or passenger_dest_y == obs[6]) and (
            partial_obs[7] == obs[7])

    def _get_passenger_info(self, partial_obs):
        passenger_start_x, passenger_start_y = partial_obs[3], partial_obs[4]
        passenger_dest_x, passenger_dest_y = partial_obs[5], partial_obs[6]
        return passenger_start_x, passenger_start_y, passenger_dest_x, passenger_dest_y


class TaxiSimpleExampleEnv(TaxiSimpleEnv):
    """
    This environment fixes the starting state for every episode to be:
    taxi starting location - (4,0)
    taxi fuel level - 3
    passenger starting location - (0, 0)
    passenger destination - (4, 4)
    """

    def reset(self) -> dict:
        """
        Reset the environment's state:
            - taxis coordinates - fixed.
            - refuel all taxis
            - destinations - fixed.
            - passengers locations - fixed.
            - preserve other definitions of the environment (collision, capacity...)
            - all engines turn on.
        Args:

        Returns: The reset state.

        """
        # reset taxis locations
        taxis_locations = [TAXI_INIT_POS]
        self.collided = np.zeros(self.num_taxis)
        self.bounded = False
        self.window_size = 5
        self.counter = 0

        # refuel everybody
        fuels = [3 for i in range(self.num_taxis)]

        # reset passengers
        passengers_start_location = self.passengers_start_fixed_locations
        passengers_destinations = self.passengers_fixed_destinations
        self.passengers_locations = [passengers_start_location[0], passengers_destinations[0]]

        # Status of each passenger: delivered (1), in_taxi (positive number>2), waiting (2)
        passengers_status = [2 for _ in range(self.num_passengers)]
        self.state = [taxis_locations, fuels, passengers_start_location, passengers_destinations, passengers_status]

        self.last_action = None
        # Turning all engines on
        self.engine_status_list = list(np.ones(self.num_taxis))

        # resetting dones
        self.dones = {taxi_id: False for taxi_id in self.taxis_names}
        self.dones['__all__'] = False
        obs = {}
        for taxi_id in self.taxis_names:
            obs[taxi_id] = self.get_observation(self.state, taxi_id)
        obs = obs[TAXI_NAME][0]
        encoded_state = self.encode(obs)
        decoded_state = self.decode(encoded_state)
        return encoded_state




from rl.agents import SARSAAgent
from rl.policy import EpsGreedyQPolicy

from src.rlpe.Agents.abstract_agent import AbstractAgent
import numpy as np
import keras
from keras.activations import relu, linear
# import lunar_lander as lander
from collections import deque
# from keras.utils import to_categorical
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Embedding, Reshape
from src.rlpe.constants import *


learning_rate = 0.001
epsilon = 1
gamma = .99
batch_size = 64
memory = deque(maxlen=1000000)
min_eps = 0.01


class HandsOnSarsaAgent(AbstractAgent):
    def __init__(self, env, timesteps_per_episode=10001):
        super().__init__(env, timesteps_per_episode)
        self.model = self._build_compile_model()
        self.num_episodes = 400
        self.evaluating = False

    def run(self) -> {str: float}:
        """
        The agent's training method.
        Returns: a dictionary - {"episode_reward_mean": __, "episode_reward_min": __, "episode_reward_max": __,
        "episode_len_mean": __}
        """
        np.random.seed(0)
        scores = []
        for i in range(self.num_episodes + 1):
            score = 0
            state = self.env.reset()
            finished = False
            if i != 0 and i % 50 == 0:
                self.model.save(".\saved_models\model_" + str(i) + "_episodes.h5")
            for j in range(3000):
                # state = np.reshape(state, (1, 8))
                if np.random.random() <= epsilon:
                    action = np.random.choice(7)
                else:
                    predict_state = np.reshape(np.array(state), (1, 1))
                    action_values = self.model.predict(predict_state)
                    action = np.argmax(action_values[0])

                if self.evaluating:
                    self.env.render()
                next_state, reward, finished, metadata = self.env.step(action)
                # next_state = np.reshape(next_state, (1, 8))
                memory.append((state, action, next_state, reward, finished))
                self.replay_experiences()
                score += reward
                state = next_state
                if finished:
                    scores.append(score)
                    print("Episode = {}, Score = {}, Avg_Score = {}".format(i, score, np.mean(scores[-100:])))
                    break

    def _build_compile_model(self):
        model = keras.Sequential()
        model.add(keras.layers.Dense(64, input_dim=1, activation=relu))
        model.add(keras.layers.Dense(64, activation=relu))
        model.add(keras.layers.Dense(1, activation=linear))
        model.compile(loss="mse", optimizer=Adam(lr=learning_rate))
        return model

    def compute_action(self, state) -> int:
        """
        Computes the best action from a given state.
        Returns: a int that represents the best action.
        """
        pass

    def stop_episode(self):
        pass

    def episode_callback(self, state, action, reward, next_state, terminated):
        pass

    def evaluate(self):
        pass

    def replay_experiences(self):
        if len(memory) >= batch_size:
            sample_choices = np.array(memory)
            mini_batch_index = np.random.choice(len(sample_choices), batch_size)
            # batch = random.sample(memory, batch_size)
            states, actions, next_states, rewards, finishes = [], [], [], [], []
            for index in mini_batch_index:
                states.append(memory[index][0])
                actions.append(memory[index][1])
                next_states.append(memory[index][2])
                rewards.append(memory[index][3])
                finishes.append(memory[index][4])
            states = np.array(states)
            actions = np.array(actions)
            next_states = np.array(next_states)
            rewards = np.array(rewards)
            finishes = np.array(finishes)
            states = np.squeeze(states)
            next_states = np.squeeze(next_states)
            q_vals_next_state = self.model.predict_on_batch(next_states)
            q_vals_target = self.model.predict_on_batch(states)
            max_q_values_next_state = np.amax(q_vals_next_state, axis=1)
            # q_vals_target[np.arange(batch_size), actions] = rewards + gamma * (max_q_values_next_state) * (1 - finishes)
            tmp = rewards + gamma * max_q_values_next_state * (1 - finishes)
            tmp = np.reshape(tmp, q_vals_target[actions].shape)
            q_vals_target[actions] = tmp
            self.model.fit(states, q_vals_target, verbose=0)
            global epsilon
            if epsilon > min_eps:
                epsilon *= 0.996


class KerasSarsaAgent(AbstractAgent):
    def __init__(self, env, timesteps_per_episode=10001):
        super().__init__(env, timesteps_per_episode)
        self.num_episodes = 400
        self.evaluating = False
        self.action_size = env.action_space.n
        self.state_size = env.num_states
        self.model = self._build_compile_model()
        self.agent = SARSAAgent(model=self.model, nb_actions=self.action_size, policy=EpsGreedyQPolicy())

    def run(self) -> {str: float}:
        """
        The agent's training method.
        Returns: a dictionary - {"episode_reward_mean": __, "episode_reward_min": __, "episode_reward_max": __,
        "episode_len_mean": __}
        """
        self.agent.compile(Adam(lr=0.001), metrics=["mse"])
        history = self.agent.fit(self.env, nb_steps=ITER_NUM, visualize=False, verbose=1)
        if len(history.history) > 0:
            episode_reward = history.history["episode_reward"]
            nb_episode_steps = history.history["nb_episode_steps"]
        else:
            episode_reward, nb_episode_steps = [0], [0]  # TODO - placeholder
        result = {EPISODE_REWARD_MEAN: np.array(episode_reward),
                  EPISODE_STEP_NUM_MEAN: np.array(nb_episode_steps),
                  EPISODE_REWARD_MIN: np.empty([]),
                  EPISODE_REWARD_MAX: np.empty([]), EPISODE_VARIANCE: np.empty([])}
        return result

    def _build_compile_model(self):
        model = Sequential()
        # model.add(Flatten(input_shape=(1, self.action_size)))
        model.add(Embedding(self.state_size, 10, input_length=1))  # 600000
        model.add(Reshape((10,)))
        model.add(Dense(24, activation='relu'))
        model.add(Dense(16, activation='relu'))
        model.add(Dense(16, activation='relu'))
        model.add(Dense(self.action_size, activation='linear'))
        return model

    def compute_action(self, state) -> int:
        """
        Computes the best action from a given state.
        Returns: a int that represents the best action.
        """
        state = np.array([[state]])
        return int(np.argmax(self.model.predict(state)))

    def stop_episode(self):
        pass

    def episode_callback(self, state, action, reward, next_state, terminated):
        pass

    def evaluate(self, visualize=False):
        self.agent.test(self.env, nb_episodes=5, visualize=visualize, nb_max_episode_steps=60)

    def replay_experiences(self):
        pass



from src.rlpe.constants import *
from src.rlpe.Agents.RL_agents import rl_agent
import numpy as np


# helper function for flattening irregular nested tuples
def mixed_flatten(x):
    result = []
    for el in x:
        if hasattr(el, "__iter__"):
            result.extend(mixed_flatten(el))
        else:
            result.append(el)
    return result


# helper function for making a list of coordinates of interest
# make list of valid coords in environment within dist of the given loc
def get_nearby_coords(env, loc, dist):  # option for later: add an option to change definition of distance
    max_rows = env.num_rows - 1
    max_cols = env.num_columns - 1
    (x, y) = loc
    result = []
    for i in range(x - dist, x + dist + 1):
        for j in range(y - dist, y + dist + 1):
            if 0 <= i <= max_rows and 0 <= j <= max_cols:
                result.append((i, j))
    return result


def sample_anticipated_policy(policy_dict, num_states_in_partial_policy):
    # get flat indices of sampled states
    sampled_states_flat = np.random.choice(len(policy_dict), size=num_states_in_partial_policy, replace=False)

    partial_sampled_policy = {}
    for i, item in enumerate(policy_dict.items()):
        if i in sampled_states_flat:
            partial_sampled_policy[tuple(item[0])] = item[1]
    return partial_sampled_policy


def is_interesting_state(state, passenger_origins, passenger_destinations):
    taxi_location = [state[0], state[1]]
    fuel_level = state[2]
    passenger_location = [state[3], state[4]]
    passenger_destination = [state[5], state[6]]
    passenger_status = state[6]

    fuel_is_full = (fuel_level == 100) or (not fuel_level)
    taxi_in_interesting_location = (
            (taxi_location[0] == passenger_location[0] and taxi_location[1] == passenger_location[1]) or (
            taxi_location[0] == passenger_destination[0] and taxi_location[1] == passenger_destination[1]))
    passenger_in_interesting_location = passenger_location in passenger_origins
    # valid_passenger_destination = ((passenger_destination[0] == passenger_location[0]) and (
    #         passenger_destination[1] == passenger_location[1]) and passenger_status > 2) or (
    #                                       passenger_destination[0] != passenger_location[0]) or (
    #                                       passenger_destination[1] != passenger_location[1])

    if fuel_is_full and taxi_in_interesting_location and passenger_in_interesting_location:  # and valid_passenger_destination:
        return True
    return False


def get_possible_passenger_origins(env):
    return env.passengers_locations


def get_possible_passenger_destinations(env):
    return env.passengers_locations


def get_automatic_anticipated_policy_from_agent(env, agent_for_policy_generator, num_of_episodes,
                                                num_states_in_partial_policy):
    """
    get automatic anticipated policy from given agent
    """

    # create agent
    agent = rl_agent.create_agent(env, agent_for_policy_generator)
    # train the agent in the environment
    train_episode_reward_mean = rl_agent.run(agent, num_of_episodes, method=TRAIN)

    policy_dict = agent.policy_dict

    automatic_anticipated_policy = sample_anticipated_policy(policy_dict, env, num_states_in_partial_policy)
    return automatic_anticipated_policy
